#include <stddef.h>

#include "modsetup.h"
#include "xscommon.h"
/* v1.1 */
#include "jediedit.h"
#include "patchack.h"
/* v1.3 */
#include "shapedit.h"

#ifdef _WIN32
#define main __main
int main(void);
void mainCRTStartup(void) { ExitProcess(main()); }
#include "win32mod.h"
#endif

/*
  2018.03.16 - first attempt to fix Dark Forces - failed
  <...this project just lays here for years without any movement...>
  2020.03.06 - second attempt - success!

  DARK.EXE
  Version 1.0 (build 1)
  Jan 23 1995 14:36:09
*/

/*
  major note for all fixes below:
  for some reason game uses (-1) as boolean (TRUE) value
*/

/* [nocd/setup] */
static fix_item dlnocdis[] = {
  {0x7500C167UL, 0xEB},
  {0x0700C168UL, 0x1B}
};

/* v1.3 [nocd] - new proper No-CD with custom levels support */
static fix_item dlnocdrk[] = {
  /* for all comments in this "C:" or "C:\" means game root drive, not exactly only the "C" drive */
  {0xEC051A0AUL, 0xE0}, /* [1] instead of JEDIPATH compare to local zero variable */
  {0x85051A0DUL, 0x84}, /* [2] jump if zero (i.e. always now) - skip whole CD check code block */
  {0x170C9DF5UL, 0x1F}, /* [3] root_directory() fix required when game run from the root folder... */
  {0x5C0C9E04UL, 0x00}, /* ...because "C:" will not work properly with opendir()... */
  {0x750C9E05UL, 0x74}, /* ...since root_directory() expects only "\" or "C:\" as root folders... */
  {0x0C0C9E06UL, 0x0E}, /* ...but now all three cases "\", "C:" and "C:\" will be accepted */
  {0x030C9E09UL, 0x02},
  {0x000C9E0AUL, 0x5C},
  {0xB80C9E0DUL, 0x80},
  {0x010C9E0EUL, 0x78},
  {0x000C9E0FUL, 0x03},
  {0x000C9E11UL, 0x74},
  {0xC30C9E12UL, 0x02},
  {0x5C105B78UL, 0x4C}, /* [4] fix double "\LFD\" to "LFD\" which will be "C:\\LFD\" otherwise... */
  {0x4C105B79UL, 0x46}, /* ...and not work with some DOS VM (works fine in DOSBox though) */
  {0x46105B7AUL, 0x44},
  {0x44105B7BUL, 0x5C},
  {0x5C105B7CUL, 0x00}
};

/* v1.3 - old No-CD code to remove */
static fix_item dlnocdel[] = {
  /* [nocd/game] always jump to "mov strEnvPath, 0" (make sure strEnvPath always zero) */
  {0x740428CAUL, 0xEB},
  /* [nocd/game] mov ebx, strEnvPath => lea ebx, strEnvPath */
  {0x8B0428E7UL, 0x8D},
  /* [nocd/game] bAllowEnvJediPath = -1 (TRUE) */
  {0x00109069UL, 0xFF},
  {0x0010906AUL, 0xFF},
  {0x0010906BUL, 0xFF},
  {0x0010906CUL, 0xFF}
};

static fix_item dlkrunin[] = {
  /* [k.run] invert run logic */
  {0x7405DAF6UL, 0x75}
};

static fix_item dlkmjump[] = {
  /* [*.jump] always */
  {0x3605DBF3UL, 0x00}, /* keyboard */
  {0x3705E749UL, 0x00}, /* mouse */
  {0x3005F002UL, 0x00}  /* joystick */
};

/* v1.4 - new mouse [Secondary Fire] */
static fix_item dlmsfire[] = {
/*
    * three consecutive null func at offset 000553F0 00055405 and 0005541A was used for mouse
      [Secondary Fire] code placing in the game executable file, since compared to mouse looking code
      it's new additional feature and nothing should be changed, replaced or removed
    * null func is a func that does nothing, but useless pushes and pops registers to / from stack
      so if func immediately returns (C3) all these wasted bytes can be used for something meaningful
      and if you wonder why this code exists in the first place - it's because compilers at that time
      didn't optimize and remove unused or obviously empty code (it's nice, since this come in handy)
    * there are 16 key binding slots for mouse in configuration file: 11 used + 4 unused and 1 under
      number 11 (which goes right before [Primary Fire]) completely ignored by the game - this slot
      will be used to add [Secondary Fire] since people confused why they need to bind [Strafe Mode]
      to right mouse button to be able to use [Secondary Fire] with old hack (*sigh*)
*/
  {0x530553F0UL, 0xC3}, /* each null func must return immediately since its body changed */
  {0x510553F1UL, 0x89}, /* [1] mov  edi, eax ; save [Primary Fire] slot address (6320) */
  {0x520553F2UL, 0xC7}, /* mov  ax,  [edi - 2] ; get ignored slot state - right before [Primary Fire] */
  {0x560553F3UL, 0x66},
  {0x570553F4UL, 0x8B},
  {0x550553F5UL, 0x47},
  {0x890553F6UL, 0xFE},
  {0xE50553F7UL, 0x31}, /* xor edx, edx */
  {0x810553F8UL, 0xD2},
  {0xEC0553F9UL, 0x66}, /* mov dx, ax ; save current slot state for later usage */
  {0x000553FAUL, 0x89},
  {0x000553FBUL, 0xC2},
  {0x000553FCUL, 0x30}, /* xor al, al */
  {0x000553FDUL, 0xC0},
  {0x5D0553FEUL, 0xF6}, /* not ah */
  {0x5F0553FFUL, 0xD4},
  {0x5E055400UL, 0x66}, /* test ax, bx ; bx contains some mask data which need to be tested */
  {0x5A055401UL, 0x85},
  {0x59055402UL, 0xD8},
  {0x5B055403UL, 0xEB}, /* jump to [2] */
  {0xC3055404UL, 0x01},
  {0x53055405UL, 0xC3}, /* each null func must return immediately since its body changed */
  {0x51055406UL, 0x75}, /* [2] jne to [4] */
  {0x52055407UL, 0x1B},
  {0x56055408UL, 0x89}, /* mov eax, edx ; restore saved slot state  */
  {0x57055409UL, 0xD0},
  {0x5505540AUL, 0x30}, /* xor ah, ah */
  {0x8905540BUL, 0xE4},
  {0xE505540CUL, 0x83}, /* cmp  dword [edi + 017ECCh + eax*4], 0 ; 6320 + 17ECC = 01E1EC */
  {0x8105540DUL, 0xBC}, /* this will check if that key was pressed */
  {0xEC05540EUL, 0x87},
  {0x0005540FUL, 0xCC},
  {0x00055410UL, 0x7E},
  {0x00055411UL, 0x01},
  {0x5D055413UL, 0x00},
  {0x5F055414UL, 0x74}, /* je to [4] ; if key wasn't pressed */
  {0x5E055415UL, 0x0D},
  {0x5A055416UL, 0x31}, /* xor eax, eax ; eax = 0 */
  {0x59055417UL, 0xC0},
  {0x5B055418UL, 0xEB}, /* jump to [3] */
  {0xC3055419UL, 0x01},
  {0x5305541AUL, 0xC3}, /* each null func must return immediately since its body changed */
  {0x5105541BUL, 0xF7}, /* [3] not eax ; eax = -1 (TRUE) */
  {0x5205541CUL, 0xD0},
  {0x5605541DUL, 0x89}, /* mov  [edi + 0C8h], eax ; 6320 + 0C8 = 63E8 ([Secondary Fire] flag) */
  {0x5705541EUL, 0x87},
  {0x5505541FUL, 0xC8},
  {0x89055420UL, 0x00},
  {0xE5055421UL, 0x00},
  {0x81055422UL, 0x00},
  {0xEC055423UL, 0x66}, /* [4] mov  ax,  [edi] ; this and next instruction emulate original code... */
  {0x00055424UL, 0x8B}, /* ...which was replaced to call this null funcs */
  {0x00055425UL, 0x07},
  {0x00055426UL, 0x25}, /* and  eax, 0FF00h */
  {0x5D055428UL, 0xFF},
  {0x5F055429UL, 0x00},
  {0x5E05542AUL, 0x00},
  {0x5A05542BUL, 0xC3}, /* retn to the original code */
  {0x6605EA8CUL, 0x8D}, /* replace "mov ax,[00006320]" with "lea eax,[00006320]" (save address)... */
  {0xA105EA8DUL, 0x05}, /* ... of [Primary Fire] slot */
  {0x3005EA92UL, 0x84}, /* xor al, al => test al, al (preserve eax address value) */
  {0x2505EA94UL, 0xE8}, /* call to [1] ... */
  {0xFF05EA95UL, 0x58}, /* ... will return here to continue with [Primary Fire] ... */
  {0xFF05EA96UL, 0x69}, /* ... and since this is the end of the mouse control function ... */
  {0x0005EA97UL, 0xFF}, /* ... edi register can be used to reduce hack code since this... */
  {0x0005EA98UL, 0xFF}, /* ... register will be restored from the stack before func return anyway */
  {0x0010AB1EUL, 0x06}, /* set ignored slot from 00 00 (completely ignored) to 06 (Right Mouse Button) */
  {0x0010AB1FUL, 0x0F}  /* and 0F (action firing or whatever it was) */
};

/* v1.4 - old mouse [Secondary Fire] code to remove */
static fix_item dlmsfdel[] = {
  /* [m.fire] */
  {0x3005E3E7UL, 0x33}, /* xor edx, edx (edx = 0) */
  {0xE405E3E8UL, 0xD2},
  {0xFF05E3EBUL, 0x00}, /* and eax, 0FFFFh => and eax, 0FFh */
  {0xC705E3F8UL, 0x8D}, /* lea eax, [1E360] */
  {0xFF05E3FEUL, 0x2D}, /* sub eax, 017F78h ( => addr of [Secondary Fire] value ) */
  {0xFF05E3FFUL, 0x78},
  {0xFF05E400UL, 0x7F},
  {0xFF05E401UL, 0x01},
  {0xBB05E402UL, 0x00},
  {0x0005E403UL, 0xF7}, /* not edx (edx = -1 (TRUE)) */
  {0x0205E404UL, 0xD2},
  {0x0005E405UL, 0x89}, /* mov [eax], edx */
  {0x0005E406UL, 0x10}
};

static fix_item dlmsensx[] = {
  /* v1.3 [m.sens/x] */
  {0x10054C60UL, 0x0F}
};

static fix_item dlmsensy[] = {
  /* v1.3 [m.sens/y] */
  {0x10054C6EUL, 0x0F}
};

static fix_item dlaimoff[] = {
  /* v1.2 [aim] */
  {0x0707A848UL, 0x00}
};

static fix_item dlmidfix[] = {
  /* v1.1 [midi] */
  /* GMIDI.MDR - General MIDI */
  {0x04029F29UL, 0x07},
  /* ROLAND.MDR - Roland */
  {0x0403470CUL, 0x07}
};

static fix_item dlmvlook[] = {
  /* [m.vert] */
  /*
    explanation for asm code for mouse code hack

    // original asm code - check that ecx (mouse Y delta value) non-zero
    test ecx, ecx
    @check:
    jz @somewhere

      // the code below in C-like notation for better understanding
      ecx >>= 5; // too big for raw use, scale down (ecx /= 32)
      eax = someaddr; // some variable address
      eax = eax - 0xA0; // actual vertical view variable addr
      ecx += *eax; // add current view value to mouse Y delta
      // next code checks that view angle inside intended [-2047..2047] range
      edx = 2047;
      if (ecx > edx) { ecx = edx; }
      edx = -edx;
      if (ecx < edx) { ecx = edx; }
      // update current view angle
      *eax = ecx;

    // back to asm code
    // jump to @check label above with zero flag set
    // so this code block will be skipped to a next part
    xor ecx, ecx ; set zero flag
    jmp @check   ; jump back
    <...a lot of code goes here...>
    @somewhere:
  */
  {0x8905E444UL, 0xF7}, /* neg ecx */
  {0xCA05E445UL, 0xD9},
  {0xF705E446UL, 0xC1}, /* sar ecx, 05 */
  {0xDA05E447UL, 0xF9},
  {0x6605E448UL, 0x05},
  {0x8905E449UL, 0x8D}, /* lea eax, [63D0] */
  {0x1D05E44AUL, 0x05},
  {0x8905E44FUL, 0xEB}, /* skip 4 bytes with reloc data */
  {0x9005E450UL, 0x04},
  {0x8505E455UL, 0x2D}, /* sub eax, 0A0h (63D0 => 6330 - addr of view value) */
  {0xD205E456UL, 0xA0},
  {0x7C05E457UL, 0x00},
  {0x7505E458UL, 0x00},
  {0xB805E459UL, 0x00},
  {0x0005E45AUL, 0x03}, /* add ecx, [eax] */
  {0x0005E45BUL, 0x08},
  {0x0005E45CUL, 0xEB}, /* skip 6 bytes with reloc data */
  {0x0105E45DUL, 0x06},
  {0xF705E464UL, 0xBA}, /* mov edx, 07FF */
  {0xEA05E465UL, 0xFF},
  {0x0F05E466UL, 0x07},
  {0xAC05E467UL, 0x00},
  {0xD005E468UL, 0x00},
  {0x1005E469UL, 0x3B}, /* cmp ecx, edx */
  {0x6605E46AUL, 0xCA},
  {0x8B05E46BUL, 0xEB}, /* skip 4 bytes with reloc data */
  {0x1505E46CUL, 0x04},
  {0x3005E471UL, 0x7E}, /* jle next instruction */
  {0xF605E472UL, 0x02},
  {0x8105E473UL, 0x8B}, /* mov ecx, edx */
  {0xE205E474UL, 0xCA},
  {0xFF05E475UL, 0xF7}, /* neg edx */
  {0xFF05E476UL, 0xDA},
  {0x0005E477UL, 0x3B}, /* cmp ecx, edx */
  {0x0005E478UL, 0xCA},
  {0x8B05E479UL, 0xEB}, /* skip 5 bytes with reloc data */
  {0x0C05E47AUL, 0x05},
  {0x8905E480UL, 0x7D}, /* jge next instruction */
  {0xC205E481UL, 0x02},
  {0x8905E482UL, 0x8B}, /* mov ecx, edx */
  {0xC805E483UL, 0xCA},
  {0xF705E484UL, 0x89}, /* mov [eax], ecx */
  {0xEA05E485UL, 0x08},
  {0x0F05E486UL, 0x31}, /* xor ecx, ecx (for contition jump) */
  {0xAC05E487UL, 0xC9},
  {0xD005E488UL, 0xEB}, /* jmps to long condition jump before this code */
  {0x1005E489UL, 0xB4}
};

static fix_item dlcrsshr[] = {
  /* v1.3 [c.hair] */
  /*
    explanation for the displaying crosshair code hack
    * four consecutive null func at offset 00062507, 0006251C, 00062531 and 00062546 was used for
      crosshair code placing in the game executable file, since compared to mouse code it's new
      additional feature and nothing should be changed, replaced or removed
    * null func is a func that does nothing, but useless pushes and pops registers to / from stack
      so if func immediately returns (C3) all these wasted bytes can be used for something meaningful
      and if you wonder why this code exists in the first place - it's because compilers at that time
      didn't optimize and remove unused or obviously empty code (it's nice, since this come in handy)
    * code below hooks to the text output func for LADATA cheat code and displays crosshair as
      custom-made font character (i.e. using already existing text output func)
    * PrnCharToScr() func prototype: PrnCharToScr(FontCharData, scr_rect, x_pos, y_pos, bank_addr);
      this func display character at position x_pos, y_pos (scr_rect and bank_addr - reserved)
    * DrawText() func prototype: DrawText(font_ptr, scr_rect, x_pos, y_pos, s_txt, bank_addr);
      this func display s_txt text at position x_pos, y_pos (font_ptr, scr_rect, bank_addr - reserved)
    * PrnCharToScr() called from DrawText(), so you can save a lot of code if you just use strings
      "." or "+" as crosshair with DrawText(), but these characters too thick with this method and
      your crosshair will be misplaced if someone changes game font file, so this code below uses
      safe approach - it creates custom-made character for crosshair and use it with PrnCharToScr()
      one of the string for text-graphic table intended for not enough memory message was used for
      placing character data structure and bitmap image there
    * in Watcom compiler func first four arguments placed in eax, edx, ebx, ecx and remains on stack
      esi and edi registers thrashed in code below since parent function restores them on exit anyway
    * the code below in C-like notation for better understanding
    // if Tab map not shown - display crosshair (it will interfere with the player's position marker)
    y = 0; // you wonder why y used as zero for all if()'s below - it's because this saves few bytes
    if (dwShowMap == y) {
      // shoot spot located at exactly (x:160,y:100)
      // with crosshair 5x5 image this will give (x:158,y:98) position where to output character
      // but needed to be moved 5 pixels up for any screen size different from maximal 320x200
      if (cfg_main.iScreenSize != y) {
        // not 320x200 - move shoot spot up by 5 pixels
        // note that this 5 not related in any kind of way to the crosshair image dimension 5x5
        y = -5;
      }
      y = y + 98; // 98 if (y == 0) or 93 if (y == -5)
      PrnCharToScr(FonCharData, scr_rect, 158, y, bank_addr);
    }
    ...overall it was hard enough to squeeze all this features even in consecutive four func code...
  */
  {0x9604E3CBUL, 0x61}, /* [parent func] do not skip text output if LADATA cheat not enabled */
  {0x0004E44FUL, 0xB5}, /* call null func [1] instead of DrawText() function */
  {0x2004E450UL, 0x40},
  {0x0304E451UL, 0x01},
  {0x53062507UL, 0xC3}, /* each null func must return immediately since its body changed */
  {0x51062508UL, 0x74}, /* [1] jz +14 - last operation non-zero - it's "add esp, 24h" after sprintf()... */
  {0x52062509UL, 0x0E}, /* ...and code below will work or will be skipped to [!] for zero */
  {0x5606250AUL, 0x5F}, /* pop edi - get from stack return address */
  {0x5706250BUL, 0x89}, /* mov esi, edx - save scr_rect argument since it will be needed later */
  {0x5506250CUL, 0xD6},
  {0x8906250DUL, 0xE8}, /* call DrawText() */
  {0xE506250EUL, 0x41},
  {0x8106250FUL, 0xDF},
  {0xEC062510UL, 0x01},
  {0x00062512UL, 0x83}, /* [2] sub esp, 8 - rewind stack after DrawText() call */
  {0x00062513UL, 0xEC},
  {0x00062514UL, 0x08},
  {0x5D062515UL, 0x57}, /* push edi - put return address back to stack */
  {0x5F062516UL, 0x89}, /* mov edx, esi - restore edx */
  {0x5E062517UL, 0xF2},
  {0x5A062518UL, 0x58}, /* [!] pop eax - get return address from stack */
  {0x59062519UL, 0xEB}, /* jump to [2] */
  {0x5B06251AUL, 0x02},
  {0x5306251CUL, 0xC3}, /* each null func must return immediately since its body changed */
  {0x5106251DUL, 0x89}, /* [2] mov [esp], eax - fix return address... */
  {0x5206251EUL, 0x04}, /* ...since this code called after DrawText() or without it and... */
  {0x5606251FUL, 0x24}, /* ...keep bank_addr on stack and replace s_txt with correct return address */
  {0x57062520UL, 0x31}, /* xor ecx, ecx (ecx == 0) */
  {0x55062521UL, 0xC9},
  {0x89062522UL, 0x39}, /* cmp [edx - 0C99h], ecx - check if dwShowMap == 0 (no Tab map on screen) */
  {0xE5062523UL, 0x8A},
  {0x81062524UL, 0x67},
  {0xEC062525UL, 0xF3},
  {0x00062526UL, 0xFF},
  {0x00062527UL, 0xFF},
  {0x00062528UL, 0x74}, /* jump to [3] if dwShowMap equal to zero... */
  {0x00062529UL, 0x08},
  {0x5D06252AUL, 0xC2}, /* retn 4 - ...or exit and do not show crosshair when map on screen */
  {0x5F06252BUL, 0x04},
  {0x5E06252CUL, 0x00},
  {0x53062531UL, 0xC3}, /* each null func must return immediately since its body changed */
  {0x51062532UL, 0x39}, /* [3] cmp [edx - 0D81h], ecx - check if cfg_main.iScreenSize == 0 (320x200 mode) */
  {0x52062533UL, 0x8A},
  {0x56062534UL, 0x7F},
  {0x57062535UL, 0xF2},
  {0x55062536UL, 0xFF},
  {0x89062537UL, 0xFF},
  {0xE5062538UL, 0x74}, /* skip code if it is 320x200 (0) */
  {0x81062539UL, 0x03},
  {0xEC06253AUL, 0x83}, /* sub ecx, 5 - in any mode less than 320x200 shoot-point moved up by 5px ! */
  {0x0006253BUL, 0xE9},
  {0x0006253CUL, 0x05},
  {0x0006253DUL, 0x8D}, /* lea eax, [edx - 4212h] - FontCharData location */
  {0x0006253EUL, 0x82},
  {0x5D06253FUL, 0xEE},
  {0x5F062540UL, 0xBD},
  {0x5E062541UL, 0xFF},
  {0x5A062542UL, 0xFF},
  {0x59062543UL, 0xEB}, /* jump to [4] */
  {0x5B062544UL, 0x02},
  {0x53062546UL, 0xC3}, /* each null func must return immediately since its body changed */
  {0x51062547UL, 0x83}, /* [4] add ecx, 98 - y_pos where draw crosshair */
  {0x52062548UL, 0xC1},
  {0x56062549UL, 0x62},
  {0x5706254AUL, 0x8D}, /* lea ebx, [eax + 1Ch] - offset to bitmap character data */
  {0x5506254BUL, 0x58},
  {0x8906254CUL, 0x1C},
  {0xE506254DUL, 0x89}, /* mov [eax + 10], ebx - fix pointer to bitmap data */
  {0x8106254EUL, 0x58},
  {0xEC06254FUL, 0x10},
  {0x00062550UL, 0xBB}, /* mov ebx, 158 - x_pos where draw crosshair */
  {0x00062551UL, 0x9E},
  {0x5D062554UL, 0x00},
  {0x5F062555UL, 0xE9}, /* jump to PrnCharToScr() - can jump since return address already on stack */
  {0x5E062556UL, 0x64},
  {0x5A062557UL, 0xCE},
  {0x59062558UL, 0x01},
  {0x5B062559UL, 0x00},
  {0xC9105BFDUL, 0x00}, /* [font data] cut one of the string to fit FontCharData here (below) */
  {0xCD105BFEUL, 0x05}, /* wWidth */
  {0xCD105BFFUL, 0x00},
  {0xCD105C00UL, 0x05}, /* wHeight */
  {0xCD105C01UL, 0x00},
  {0xCD105C02UL, 0x00}, /* dwFlag1 */
  {0xCD105C03UL, 0x00},
  {0xCD105C04UL, 0x00},
  {0xCD105C05UL, 0x00},
  {0xCD105C06UL, 0x19}, /* dwSize = wWidth * wHeight */
  {0xCD105C07UL, 0x00},
  {0xCD105C08UL, 0x00},
  {0xCD105C09UL, 0x00},
  {0xCD105C0AUL, 0x00}, /* dwFlag2 and after it lpDataPtr (not present here since filled by [4]) */
  {0xCD105C0BUL, 0x00},
  {0xCD105C0CUL, 0x00},
  {0xCD105C0DUL, 0x00},
  {0xCD105C12UL, 0x00}, /* dwFlag3 */
  {0xCD105C13UL, 0x00},
  {0xCD105C14UL, 0x00},
  {0xCD105C15UL, 0x00},
  {0xCD105C16UL, 0x09}, /* bFlag4 */
  {0xCD105C17UL, 0x00}, /* bFlag5 */
  {0xCD105C18UL, 0x00}, /* wPadding */
  {0xCD105C19UL, 0x00},
  {0xCD105C1AUL, 0x00}, /* from here goes character image for the crosshair 5x5 pixels */
  {0xCD105C1BUL, 0x00},
  {0xCD105C1CUL, 0x15},
  {0xCD105C1DUL, 0x00},
  {0xCD105C1EUL, 0x00},
  {0xCD105C1FUL, 0x00},
  {0xCD105C20UL, 0x00},
  {0xCD105C21UL, 0x00},
  {0xCD105C22UL, 0x00},
  {0xCD105C23UL, 0x00},
  {0xCD105C24UL, 0x15},
  {0xCD105C25UL, 0x00},
  {0xCD105C26UL, 0x00},
  {0xCD105C27UL, 0x00},
  {0xCD105C28UL, 0x15},
  {0xCD105C29UL, 0x00},
  {0xCD105C2AUL, 0x00},
  {0xCD105C2BUL, 0x00},
  {0xCD105C2CUL, 0x00},
  {0xCD105C2DUL, 0x00},
  {0xCD105C2EUL, 0x00},
  {0xCD105C2FUL, 0x00},
  {0xCD105C30UL, 0x15},
  {0xCD105C31UL, 0x00},
  {0xCD105C32UL, 0x00}
};

/* v1.3 mouse linear scale */
static fix_item dlmlnscl[] = {
  /* replace hardcoded static table for x_sens, y_sens, m_boost with linear one
     note that only x_sens and y_sens will be replaced */
  {0x1A04A80FUL, 0x0C},
  {0x2304A811UL, 0x1E},
  {0x1B04A812UL, 0x12},
  {0x3704A814UL, 0x28},
  {0x1C04A815UL, 0x19},
  {0x4104A817UL, 0x32},
  {0x1D04A818UL, 0x1F},
  {0x4604A81AUL, 0x3C},
  {0x1E04A81BUL, 0x25},
  {0x4B04A81DUL, 0x46},
  {0x1F04A81EUL, 0x2B},
  {0x2004A821UL, 0x32},
  {0x5504A823UL, 0x5A},
  {0x2104A824UL, 0x38},
  {0x2204A827UL, 0x3E},
  {0x0705D64EUL, 0x06}, /* equal shift 6 for both axes (axis Y already shifted) */
  {0x0705E3B4UL, 0x06}
};

static fix_item dlmsgtxt[] = {
  /* v1.4 [m.text] */
  {0x5D04E0A8UL, 0x00}
};

static fix_item dlwdlock[] = {
  /* v1.4 [w.lock] */
  {0x89060AECUL, 0x85}
};

static fix_file filelist[] = {
  /* 0 */ {"DARK.EXE", 1172219UL},
  /* 1 */ {"SETUP.EXE",  92475UL},
  /* v1.1 [midi] */
  /* 2 */ {"IMUSE.EXE", 585827UL}
};

/* v1.1
   stat file size list */
static fix_info infolist[] = {
  /* 00 - [nocd] SETUP.EXE */
  {PH_ORG, 1, LIST_LEN(dlnocdis), dlnocdis},
  /* 01 - [nocd] */
  {PH_ORG, 0, LIST_LEN(dlnocdrk), dlnocdrk},
  /* 02 - [k.run] */
  {PH_ORG, 0, LIST_LEN(dlkrunin), dlkrunin},
  /* 03 - [*.jump] */
  {PH_ORG, 0, LIST_LEN(dlkmjump), dlkmjump},
  /* 04 - [m.fire] */
  {PH_ORG, 0, LIST_LEN(dlmsfire), dlmsfire},
  /* 05 - v1.2 [aim] */
  {PH_ORG, 0, LIST_LEN(dlaimoff), dlaimoff},
  /* 06 - v1.3 [c.hair] */
  {PH_ORG, 0, LIST_LEN(dlcrsshr), dlcrsshr},
  /* 07 - v1.3 [m.lscl] */
  {PH_ORG, 0, LIST_LEN(dlmlnscl), dlmlnscl},
  /* 08 - [m.sens/x] */
  {PH_ORG, 0, LIST_LEN(dlmsensx), dlmsensx},
  /* 09 - [m.sens/y] */
  {PH_ORG, 0, LIST_LEN(dlmsensy), dlmsensy},
  /* 10 - [m.vert] */
  {PH_ORG, 0, LIST_LEN(dlmvlook), dlmvlook},
  /* 11 - v1.4 [m.text] */
  {PH_ORG, 0, LIST_LEN(dlmsgtxt), dlmsgtxt},
  /* 12 - v1.4 [w.lock] */
  {PH_ORG, 0, LIST_LEN(dlwdlock), dlwdlock},
  /* 13 - v1.1 [midi] IMUSE.EXE */
  {PH_ORG, 2, LIST_LEN(dlmidfix), dlmidfix}
};

/* v1.3 bad, removed and superseded things to safe rollback */
static fix_info backlist[] = {
  /* old improper No-CD before v1.3 */
  {PH_ORG, 0, LIST_LEN(dlnocdel), dlnocdel},
  /* v1.4 old mouse [Secondary Fire] hack which replaces [Strafe Mode] */
  {PH_ORG, 0, LIST_LEN(dlmsfdel), dlmsfdel}
};

/* v1.3 crosshair */
static fix_info crhrlist[] = {
  {PH_ORG, 0, 5 * 5, &dlcrsshr[LIST_LEN(dlcrsshr) - (5 * 5)]}
};

static menuitem mainmenu[] = {
  {
    /* menu title */
    "Main menu",
    /* subheader (green) */
    NULL,
    /* total menu items (this not included) */
    7
  },
  {
    /* menu item name */
    "Hacks setup",
    /* menu item hint */
    "Configure various hacks.",
    /* menu item flag */
    0
  },
  /* v1.1 */
  {
    "Controls setup",
    "Change key and button assignments and a few other settings in \"JEDI.CFG\" file.",
    0
  },
  /* v1.4 */
  {
    "Set Modern Controls",
    "Automatically set Default Modern Controls and Hacks.",
    0
  },
  {
    "Advertisements",
    "Check out these related Advertisements!",
    0
  },
  {
    "About DeHacker",
    "Information about this program.",
    0
  },
  {
    "Save settings and Exit",
    "Save the current settings and Exit.",
    0
  },
  {
    "Exit without saving",
    "Exit without saving any changes.",
    0
  }
};

/* v1.3 [m.sens] */
static menuitem mainsens[] = {
  {"Select sensitivity increase number", NULL, 3},
  {"<disabled>", NULL, 0},
  {"by 2 times", NULL, 0},
  {"by 4 times", NULL, 0}
};

/* v1.4 compact [m.vert] hack */
static menuitem mainlook[] = {
  {"Select mouse look Y-axis type", NULL, 3},
  {"<disabled>", NULL, 0},
  {"Y standard", NULL, 0},
  {"Y inverted", NULL, 0}
};

static menuitem mainctrl[] = {
  {"Hacks setup", NULL, 13},
  {
    "Disable CD checks (No-CD)",
    "The game and setup program will work without a CD.",
    ITEM_CHKBOX
  },
  {
    "Enable auto-run (hold [Run] button to walk)",
    "Always run when the [Run] keyboard key is NOT pressed. Makes gameplay harder.",
    ITEM_CHKBOX
  },
  {
    "Keep jumping while the [Jump] button is pressed",
    "No need to hit [Jump] key continuously. May help with mine or detonator jumps.",
    ITEM_CHKBOX
  },
  {
    "Enable [Secondary Fire] on mouse",
    "To use alt-fire mode with mouse setup's [Secondary Fire] as a firing button.",
    ITEM_CHKBOX
  },
  /* v1.2 [aim] */
  {
    "Disable weapon aim-assist",
    "Auto-aiming assistance will be disabled for all weapons.",
    ITEM_CHKBOX
  },
  /* v1.3 [c.hair] */
  {
    "Enable weapon crosshair",
    "Very useful without aim-assist. Note: Only the blaster has high firing accuracy.",
    ITEM_CHKBOX
  },
  /* v1.3 [m.lscl] */
  {
    "Enable better mouse sensitivity scale for both axes",
    "Replaces mouse sensitivity table with linear axes scalings.",
    ITEM_CHKBOX
  },
  /* v1.3 [m.sens/x] */
  {
    "Increase X-axis mouse sensitivity: <disabled>",
    "Mouse X-axis sensitivity will be increased by the selected multiplier.",
     ITEM_CHKBOX
  },
  /* v1.3 [m.sens/y] */
  {
    "Increase Y-axis mouse sensitivity: <disabled>",
    "Mouse Y-axis sensitivity will be increased by the selected multiplier.",
     ITEM_CHKBOX
  },
  /* v1.4 compact [m.vert] hack */
  {
    "Enable vertical mouse look: <disabled>",
    "Allow vertical mouse look, [Move / Look Up/Down] for [Y Axis] is required.",
    ITEM_CHKBOX
  },
  /* v1.4 [m.text] */
  {
    "Disable notification messages priority",
    "New messages will always overwrite old ones when you pickup or use something.",
    ITEM_CHKBOX
  },
  /* v1.4 [w.lock] */
  {
    "Lock vertical weapon offset",
    "May help if you have nausea. Also don't forget to disable Head Wave with F6 key.",
    ITEM_CHKBOX
  },
  /* v1.1 [midi] */
  {
    "Enable IMUSE music MIDI drivers fix",
    "Use this if the game or sound setup hangs for General MIDI / Roland music cards.",
    ITEM_CHKBOX
  }
};

static menuitem maininfo[] = {
  {"About DeHacker", NULL, 9 | MENU_NFOBOX},
  {"Dark Forces DeHacker Version 1.4"
  #ifndef _WIN32
  " [DOS16]"
  #else
  " [Win32]"
  #endif
  , NULL, 0},
  {"(c) -=CHE@TER=- 2020,2021", NULL, 0},
  {"e-mail: _CTPAX_@MAIL.RU", NULL, 0},
  {"http://ctpax-cheater.losthost.org/htmldocs/trouble.htm", NULL, 0},
  {"Many thanks to Taton Qek, Syrion and Karjala for testing!", NULL, 0},
  {"", NULL, 0},
  {"Compatibility fixes for DOS and old Windows games and software.", NULL, 0},
  {"Hiring? Offering a one-time job? Paid requests are always welcome!", NULL, 0},
  {"Contact for more details.", NULL, 0}
};

static menuitem notfound[] = {
  {"Files not found", NULL, 3 | MENU_NFOBOX},
  {"Supported game executable not found.", NULL, 0},
  {"Move this program to the game folder where \"DARK.EXE\" file located and", NULL, 0},
  {"run again. Note that only full retail version of the game supported.", NULL, 0}
};

static menuitem mainmisc[] = {
  {"Advertisement", NULL, 17 | MENU_NFOBOX},
  {"Introducing Reviver, a patch for the game Scorcher (1996, Zyrinx/Scavenger).", NULL, 0},
  {"Some of the Scorcher Reviver features:", NULL, 0},
  {"* support for retro gaming PCs and modern systems - you can run Reviver", NULL, 0},
  {"  on anything from Windows 98 to Windows 10 or even under *nix with Wine", NULL, 0},
  {"* screen resolution not limited to 640x480 anymore, fullscreen and", NULL, 0},
  {"  new windowed borderless fullscreen mode", NULL, 0},
  {"* fully configurable gamepad / joystick / game controller, usable in menu", NULL, 0},
  {"* new challenging game modes: Survival, Insanity and Speedrun", NULL, 0},
  {"* streaming music from wave-files - CD disc not required to play", NULL, 0},
  {"* fixes to improve gameplay experience such as auto use 1st person view", NULL, 0},
  {"* a lot of original bugs patched like non saved highscores and settings", NULL, 0},
  {"* partial modding support with external files and much, much more!", NULL, 0},
  {"Try it now with the demo version of the game:", NULL, 0},
  {"https://games.softpedia.com/get/Patch/Scorcher-Reviver.shtml", NULL, 0/*ITEM_GRAYED*/},
  {"Or win a free registration key for Reviver (see topic page 37 for details):", NULL, 0},
  {"https://forum.speeddemosarchive.com/post/cash_bounties_for_speedruns361.html", NULL, 0/*ITEM_GRAYED*/},
  {"(c) SysTools 2018-2020 Reviver is only a patch, game itself required to run.", NULL, 0}
};

/* v1.4 */
static menuitem mainaskm[] = {
  {"This is your first time running DeHacker",
   "Please choose a configuration", 3},
  {"Set crosshair, mouselook and keybind controls. Save and exit (Recommended)", NULL, 0},
  {"Set crosshair, mouselook and keybind controls and let me make changes", NULL, 0},
  {"Let me customize my own controls and configuration (Advanced)", NULL, 0}
};

/* v1.4 */
static menuitem maincnfm[] = {
  {"Set Modern Controls?", NULL, 2 | MENU_DLGBOX},
  {"Yes", NULL, 0},
  {"No", NULL, 0}
};

void ApplyModernSetup(void) {
int i, n, r;
  /* hack items count */
  n = mainctrl[0].flag & MENU_ITEMSK;
  /* hacks mask */
  r = 0x7E8;
  /* apply hacks */
  for (i = 1; i <= n; i++) {
    /* item must be selected and enabled (IMUSE hack) */
    if ((r & 1) && (!ITS_GRAY(mainctrl[i]))) {
      SET_LITE(mainctrl[i], 1);
      if ((i >= 8) && (i <= 10)) {
        /* set menu data */
        SetMenuData(mainctrl[i], 1);
        /* update menu text */
        StrOverlap(
          &mainctrl[i].item[(i == 10) ? 28 : 35],
          ((i == 10) ? mainlook : mainsens)[GetMenuData(mainctrl[i]) + 1].item
        );
      }
    } else {
      /* disable if something was on */
      SET_LITE(mainctrl[i], 0);
    }
    r >>= 1;
  }
  /* something changed */
  SetMenuData(mainmenu[1], 1);
  /* save default configuration */
  MakeModernConfig();
}

int CheckAskModern(void) {
int i, n, r;
  /* check if any hacks selected */
  r = 1;
  n = mainctrl[0].flag & MENU_ITEMSK;
  for (i = 1; i <= n; i++) {
    /* something already selected */
    if (ITS_LITE(mainctrl[i])) {
      r = 0;
      break;
    }
  }
  /* if nothing was selected it's probably first time running -
     ask about apply default modern configuration */
  if (r) {
    r = DrawMenu(mainaskm, NULL);
    /* something was selected */
    if ((r == 1) || (r == 2)) {
      ApplyModernSetup();
    } else {
      r = 0;
    }
  }
  return(r);
}

/* v1.3 [m.sens/x/y] and [c.hair], v1.4 [m.vert] */
int HackProc(int item) {
menuitem *m; /* v1.4 */
int i;
  /* X/Y axis sensitivity, v1.4 compact [m.vert] hack */
  if ((item >= 8) && (item <= 10)) {
    m = (item == 10) ? mainlook : mainsens; /* v1.4 */
    DefMenuItem(m, GetMenuData(mainctrl[item]));
    i = DrawMenu(m, NULL);
    /* new sensitivity selected */
    if (i) {
      /* set new value */
      SetMenuData(mainctrl[item], i - 1);
      /* change menu item text */
      StrOverlap(&mainctrl[item].item[(item == 10) ? 28 : 35], m[i].item); /* v1.4 */
    }
    /* get current value */
    i = GetMenuData(mainctrl[item]);
    /* change state */
    SET_LITE(mainctrl[item], i);
    /* force menu redraw */
    item = -1;
  }
  /* crosshair */
  if (item == 6) {
    /* edit crosshair */
    i = EditCrosshair(crhrlist->list);
    if (i) {
      /* on or off selected */
      i = (i == 1) ? 1 : 0;
    } else {
      /* revet back if menu just cancelled */
      i = ITS_LITE(mainctrl[item]) ? 0 : 1;
    }
    /* change state */
    SET_LITE(mainctrl[item], i);
    /* force menu redraw */
    item = -1;
  }
  return(item);
}

int MenuProc(int item) {
int r;
  r = 1;
  switch (item) {
    /* hacks */
    case 1:
      /* v1.4 something changed */
      if (DrawTest(mainctrl, HackProc)) {
        SetMenuData(mainmenu[1], 1);
      }
      break;
    /* v1.1 edit */
    case 2:
      EditJediConfig(ITS_LITE(mainctrl[4])); /* v1.4 */
      break;
    /* v1.4 Modern */
    case 3:
      /* v1.4 ask only once per run about modern default configuration */
      if (DrawMenu(maincnfm, NULL) == 1) {
        ApplyModernSetup();
      }
      break;
    /* misc */
    case 4:
      DrawMenu(mainmisc, NULL);
      break;
    /* about */
    case 5:
      DrawMenu(maininfo, NULL);
      break;
    /* exit */
    case 6:
    case 7:
      r = 0;
      break;
  }
  return(r);
}

/* v1.1 */
void LoadData(void) {
int i, n;
  /* load patches state */
  LoadPatches(filelist, infolist, LIST_LEN(infolist));
  /* v1.3 safe check */
  BackPatches(filelist, backlist, LIST_LEN(backlist));
  /* v1.3 activate new No-CD if old No-CD code found */
  if (backlist[0].stat == PH_MOD) {
    infolist[1].stat = PH_MOD;
  }
  /* v1.4 activate new [Secondary Fire] if old code found  */
  if (backlist[1].stat == PH_MOD) {
    infolist[4].stat = PH_MOD;
  }
  /* v1.3 load actual crosshair shape only if crosshair enabled */
  if (infolist[6].stat == PH_MOD) {
    ReadPatches(filelist, crhrlist, LIST_LEN(crhrlist));
  }
  n = mainctrl[0].flag & MENU_ITEMSK;
  /* [*.main] */
  for (i = 1; i <= n; i++) {
    /* [midi] - last patch */
    if ((i == n) && (infolist[i].stat == PH_NON)) {
      SET_GRAY(mainctrl[i], 1);
    } else {
      SET_LITE(mainctrl[i], (infolist[i].stat != PH_ORG));
    }
    /* v1.3 [m.sens/x/y], v1.4 compact [m.vert] hack */
    if ((i >= 8) && (i <= 10)) {
      /* set menu data */
      SetMenuData(mainctrl[i], infolist[i].stat);
      /* update menu text */
      StrOverlap(
        &mainctrl[i].item[(i == 10) ? 28 : 35],
        ((i == 10) ? mainlook : mainsens)[GetMenuData(mainctrl[i]) + 1].item
      ); /* v1.4 */
    }
  }
}

void SaveData(void) {
int i, n;
  n = mainctrl[0].flag & MENU_ITEMSK;
  /* [*.main] */
  for (i = 1; i <= n; i++) {
    /* v1.4 check for [midi] */
    if (!ITS_GRAY(mainctrl[i])) {
      infolist[i].stat = ITS_LITE(mainctrl[i]) ? PH_MOD : PH_ORG;
      /* v1.3 [m.sens/x/y] */
      if ((i == 8) || (i == 9)) {
        /* 0x10 - original shift right value */
        infolist[i].list[0].data = 0x10 - GetMenuData(mainctrl[i]);
      }
      /* v1.4 [m.vert] - invert on Y axis */
      if ((i == 10) && (GetMenuData(mainctrl[i]) == 2)) {
        /* neg ecx (F7 D9) => or al, 0D9h (0C D9)
           this required to keep ecx register sign */
        infolist[10].list[0].data = 0x0C;
      }
    }
  }
  /* [nocd] SETUP.EXE */
  infolist[0].stat = infolist[1].stat;
  /* v1.3 safe rollback */
  BackPatches(filelist, backlist, LIST_LEN(backlist));
  /* save patches */
  SavePatches(filelist, infolist, LIST_LEN(infolist));
}

#ifdef _WIN32
static menuitem mainpath[] = {
  {"Select Dark Forces location", NULL, 3 | MENU_DLGBOX},
  {"Steam", NULL, 0},
  {"GOG Galaxy", NULL, 0},
  {"Current folder", NULL, 0}
};

static menuitem mainedit[] = {
  {"Error: no write permission!", NULL, 6 | MENU_NFOBOX},
  {"Game executable found but write access denied. Please make sure that:", NULL, 0},
  {"* On Windows systems DeHacker running from the administrator account", NULL, 0},
  {"  (right click on the DeHacker executable, select \"Run as Administrator\").", NULL, 0},
  {"* Game executable file doesn't have READ-ONLY attribute set (right click", NULL, 0},
  {"  on the game \"DARK.EXE\" file, select \"Properties\", uncheck \"Read-Only\").", NULL, 0},
  {"* You have a write access to the folder where the game files located.", NULL, 0}
};
#endif

/* v1.4 */
int DetectVersion(void) {
#ifndef _WIN32
  return(FileExists(filelist[0].name, filelist[0].size));
#else
int i, r;
  r = FindGameFile(filelist[0].name, filelist[0].size, 0);
  /* got something */
  if (r) {
    for (i = 0; i < 3; i++) {
      SET_GRAY(mainpath[1 + i], ((r & (1 << (2 - i))) ? 0 : 1));
    }
    do {
      i = DrawMenu(mainpath, NULL);
      if (i) {
        i = 4 - i; /* reverse */
        /* writable */
        if (r & (1 << (3 + i - 1))) {
          r = i;
          i = 0;
        } else {
          DrawMenu(mainedit, NULL);
        }
      } else {
        r = 0;
      }
    } while(i);
    if (r) {
      /* change path */
      FindGameFile(filelist[0].name, filelist[0].size, 1 << (r - 1));
    }
  }
  return(r);
#endif
}

int main(void) {
char s[34];
int i;
  i = 0;
  /* startup */
  InitSetup(maininfo[1].item, maininfo[2].item);
  /* if no game executable found */
  if (!DetectVersion()) {
    SET_GRAY(mainmenu[1], 1); /* disable hacks menu */
    SET_GRAY(mainmenu[3], 1); /* v1.4 disable modern controls */
    SET_GRAY(mainmenu[6], 1); /* disable save changes menu */
    /* show error message */
    DrawMenu(notfound, NULL);
  } else {
    /* generate menu subheader */
    StrMerge(s, 34, "Current Filename ( ", filelist[0].name, " )", NULL);
    mainmenu[0].hint = s;
    /* load patches */
    LoadData();
    /* v1.4 ask for Modern configuration */
    i = CheckAskModern();
  }
  /* v1.4 - save and exit */
  if (i != 1) {
    /* show main menu */
    do {
      i = DrawMenu(mainmenu, MenuProc);
    } while (!i); /* v1.4 prevent accidentally exit */
  } else {
    /* autosave and exit */
    i = 6;
  }
  /* if save changes or yes in additional menu */
  if (
    (GetMenuData(mainmenu[1])) && /* v1.1 any hacks changed */
    (!ITS_GRAY(mainmenu[1])) && (i == 6)
  ) {
    /* save patches */
    SaveData();
  }
  /* cleanup */
  FreeSetup();
  return(0);
}
